@@ -1,4 +1,4 @@
-# Copyright (c) Streamlit Inc. (2018-2022) Snowflake Inc. (2022-2025)
+# Copyright (c) Streamlit Inc. (2018-2022) Snowflake Inc. (2022-2026)
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -40,6 +40,7 @@ from streamlit.elements.lib.utils import (
     get_label_visibility_proto_value,
     to_key,
 )
+from streamlit.errors import StreamlitAPIException
 from streamlit.proto.Common_pb2 import FileUploaderState as FileUploaderStateProto
 from streamlit.proto.Common_pb2 import UploadedFileInfo as UploadedFileInfoProto
 from streamlit.proto.FileUploader_pb2 import FileUploader as FileUploaderProto
@@ -176,6 +177,7 @@ class FileUploaderMixin:
         args: WidgetArgs | None = None,
         kwargs: WidgetKwargs | None = None,
         *,
+        max_upload_size: int | None = None,
         disabled: bool = False,
         label_visibility: LabelVisibility = "visible",
         width: WidthWithoutContent = "stretch",
@@ -195,6 +197,7 @@ class FileUploaderMixin:
         args: WidgetArgs | None = None,
         kwargs: WidgetKwargs | None = None,
         *,
+        max_upload_size: int | None = None,
         disabled: bool = False,
         label_visibility: LabelVisibility = "visible",
         width: WidthWithoutContent = "stretch",
@@ -219,6 +222,7 @@ class FileUploaderMixin:
         on_change: WidgetCallback | None = None,
         args: WidgetArgs | None = None,
         kwargs: WidgetKwargs | None = None,
+        max_upload_size: int | None = None,
         disabled: bool = False,
         label_visibility: LabelVisibility = "visible",
         width: WidthWithoutContent = "stretch",
@@ -238,6 +242,7 @@ class FileUploaderMixin:
         on_change: WidgetCallback | None = None,
         args: WidgetArgs | None = None,
         kwargs: WidgetKwargs | None = None,
+        max_upload_size: int | None = None,
         disabled: bool = False,
         label_visibility: LabelVisibility = "visible",
         width: WidthWithoutContent = "stretch",
@@ -255,14 +260,18 @@ class FileUploaderMixin:
         args: WidgetArgs | None = None,
         kwargs: WidgetKwargs | None = None,
         *,  # keyword-only arguments:
+        max_upload_size: int | None = None,
         disabled: bool = False,
         label_visibility: LabelVisibility = "visible",
         width: WidthWithoutContent = "stretch",
     ) -> UploadedFile | list[UploadedFile] | None:
         r"""Display a file uploader widget.
+
         By default, uploaded files are limited to 200 MB each. You can
-        configure this using the ``server.maxUploadSize`` config option. For
-        more information on how to set config options, see |config.toml|_.
+        configure this globally using the ``server.maxUploadSize`` configuration
+        option. For more information on how to set configuration options, see
+        |config.toml|_. Additionally, you can set a per-widget limit using the
+        ``max_upload_size`` parameter.
 
         .. |config.toml| replace:: ``config.toml``
         .. _config.toml: https://docs.streamlit.io/develop/api-reference/configuration/config.toml
@@ -277,9 +286,9 @@ class FileUploaderMixin:
             the font height.
 
             Unsupported Markdown elements are unwrapped so only their children
-            (text contents) render. Display unsupported elements as literal
-            characters by backslash-escaping them. E.g.,
-            ``"1\. Not an ordered list"``.
+            (text contents) render. Common block-level Markdown (headings,
+            lists, blockquotes) is automatically escaped and displays as
+            literal text in labels.
 
             See the ``body`` parameter of |st.markdown|_ for additional,
             supported Markdown directives.
@@ -308,6 +317,14 @@ class FileUploaderMixin:
                 or type extensions. The correct handling of uploaded files is
                 part of the app developer's responsibility.
 
+        max_upload_size : int or None
+            The maximum allowed size of each uploaded file in megabytes.
+
+            If this is ``None`` (default), the maximum file size is set by the
+            ``server.maxUploadSize`` configuration option in your
+            ``config.toml`` file. If this is an integer, it must be positive
+            and will override the ``server.maxUploadSize`` configuration option.
+
         accept_multiple_files : bool or "directory"
             Whether to accept more than one file in a submission. This can be one
             of the following values:
@@ -322,10 +339,26 @@ class FileUploaderMixin:
             a list and a user can additively select files if they click the
             browse button on the widget multiple times.
 
-        key : str or int
-            An optional string or integer to use as the unique key for the widget.
-            If this is omitted, a key will be generated for the widget
-            based on its content. No two widgets may have the same key.
+        key : str, int, or None
+            An optional string or integer to use as the unique key for
+            the widget. If this is ``None`` (default), a key will be
+            generated for the widget based on the values of the other
+            parameters. No two widgets may have the same key. Assigning
+            a key stabilizes the widget's identity and preserves its
+            state across reruns even when other parameters change.
+
+            .. note::
+               Changing ``type``, ``accept_multiple_files``, or
+               ``max_upload_size`` resets the widget even when a key is
+               provided.
+
+            A key lets you access the widget's value via
+            ``st.session_state[key]`` (read-only). For more details, see
+            `Widget behavior
+            <https://docs.streamlit.io/develop/concepts/architecture/widget-behavior>`_.
+
+            Additionally, if ``key`` is provided, it will be used as a
+            CSS class name prefixed with ``st-key-``.
 
         help : str or None
             A tooltip that gets displayed next to the widget label. Streamlit
@@ -441,6 +474,7 @@ class FileUploaderMixin:
         return self._file_uploader(
             label=label,
             type=type,
+            max_upload_size=max_upload_size,
             accept_multiple_files=accept_multiple_files,
             key=key,
             help=help,
@@ -464,6 +498,7 @@ class FileUploaderMixin:
         args: WidgetArgs | None = None,
         kwargs: WidgetKwargs | None = None,
         *,  # keyword-only arguments:
+        max_upload_size: int | None = None,
         label_visibility: LabelVisibility = "visible",
         disabled: bool = False,
         ctx: ScriptRunContext | None = None,
@@ -471,6 +506,16 @@ class FileUploaderMixin:
     ) -> UploadedFile | list[UploadedFile] | None:
         key = to_key(key)
 
+        # Validate max_upload_size early to provide a clear error message
+        if max_upload_size is not None and (
+            not isinstance(max_upload_size, int) or max_upload_size <= 0
+        ):
+            raise StreamlitAPIException(
+                "The `max_upload_size` parameter must be a positive integer "
+                "representing the maximum file size in megabytes, or None "
+                "to fall back to the `server.maxUploadSize` configuration option."
+            )
+
         check_widget_policies(
             self.dg,
             key,
@@ -483,7 +528,12 @@ class FileUploaderMixin:
         element_id = compute_and_register_element_id(
             "file_uploader",
             user_key=key,
-            key_as_main_identity=False,
+            max_upload_size=max_upload_size,
+            # Treat the provided key as the main identity; only include changes
+            # to the type, accept_multiple_files, and max_upload_size parameters
+            # in the identity computation as those can invalidate the current
+            # value
+            key_as_main_identity={"type", "accept_multiple_files", "max_upload_size"},
             dg=self.dg,
             label=label,
             type=type,
@@ -500,9 +550,13 @@ class FileUploaderMixin:
         file_uploader_proto.type[:] = (
             normalized_type if normalized_type is not None else []
         )
-        file_uploader_proto.max_upload_size_mb = config.get_option(
-            "server.maxUploadSize"
-        )
+        if max_upload_size is not None:
+            file_uploader_proto.max_upload_size_mb = max_upload_size
+        else:
+            file_uploader_proto.max_upload_size_mb = config.get_option(
+                "server.maxUploadSize"
+            )
+
         # Handle directory uploads - they should enable multiple files and set the directory flag
         is_directory_upload = accept_multiple_files == "directory"
         file_uploader_proto.multiple_files = (
